/* Subroutines for insn-output.c for GNU compiler.  OpenRISC 1000 version.
   Copyright (C) 1987, 1992, 1997 Free Software Foundation, Inc
   Contributed by Damjan Lampret <lampret@opencores.org> in 1999.
   Based upon the elxsi (which was based upon the vax port) and sparc port.

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include "config.h"
#include <stdio.h>
#include "rtl.h"
#include "tree.h"
#include "flags.h"
#include "regs.h"
#include "hard-reg-set.h"
#include "output.h"

/* Set thist to nonzero if you want l.nop instruction in delay slot
   of l.jr instruction in epilogue. */
#define  NOP_DELAY_SLOT_FILL 0
/*
 * for holding comparison operands.
 */
rtx or1k_cmp_op0, or1k_cmp_op1;
enum machine_mode or1k_cmp_mode;

/*
 * used in function prologue/epilogue generation
 */
extern int leaf_function;

int
print_operand_punct_valid_p(code)
     int code;
{
  switch(code)
    {
      case '(':   /* idea taken from sparc; output nop for %( if
                     not optimizing or the slot is not filled. */
      case '%':
        return 1;
    }
  return 0;
}
                                                  
print_operand_address (file, addr)
     FILE *file;
     register rtx addr;
{
  register rtx reg1, reg2, breg, ireg;
  rtx offset;
 
 retry:
  switch (GET_CODE (addr))
    {

    case MEM:
      if (GET_CODE (XEXP (addr, 0)) == REG)
        fprintf (file, "%s", reg_names[REGNO (addr)]);
      else abort();
      break;

    case REG:
      fprintf (file, "0(%s)", reg_names[REGNO (addr)]);
      break;

    case PLUS:
      reg1 = 0;	reg2 = 0;
      ireg = 0;	breg = 0;
      offset = 0;
      if (GET_CODE (XEXP (addr, 0)) == REG)
	{
	  offset = XEXP (addr, 1);
	  addr = XEXP (addr, 0);
	}
      else if (GET_CODE (XEXP (addr, 1)) == REG)
	{
	  offset = XEXP (addr, 0);
	  addr = XEXP (addr, 1);
	}
      output_address (offset);
      fprintf (file, "(%s)", reg_names[REGNO (addr)]);
      break;

    default:
      /* fprintf(file, "{%d}", GET_CODE (addr)); */
      output_addr_const (file, addr);
    }
}

char *
get_function_name()
{
  char *kind = "function";
  if (current_function_decl != 0
      && TREE_CODE (TREE_TYPE (current_function_decl)) == METHOD_TYPE)
    kind = "method";

  if (current_function_decl == NULL)
    {
      return "top level";
    }
  else
    {
      char *name = (*decl_printable_name) (current_function_decl, &kind);
      return name;
    }
}

/*
 * operands[0] will be the branch target.
 * test will be the comparison type.
 * or1k_cmp_mode is SImode
 * or1k_cmp_op0 and or1k_cmp_op1 are the things to be 
 * compared.
 *
 * This routine has to generate a compare and a branch rtx.
 *
 */
int
gen_cond_branch(operands, test)
     rtx operands[];
     enum rtx_code test;
{
  rtx tmp;

  switch(or1k_cmp_mode)
    {
    case SImode:
      /* generate a sfXX instruction. */

      /* check if operand 1 is an integer that is too big;
         move it into a register if it is. */

#if 1 /* Set to 0 if sfXXi are not available (also cmpsi should compare only registers (no imms)) */
      if(GET_CODE(or1k_cmp_op1) == CONST_INT)
        {
          int c = INTVAL(or1k_cmp_op1);
/*          if (!(CONST_OK_FOR_LETTER_P(c, 'I')))
             debug_rtx(or1k_cmp_op1); */
             or1k_cmp_op1 = force_reg(SImode, or1k_cmp_op1);
        }
#endif
      else if(GET_CODE(or1k_cmp_op1) == SYMBOL_REF)
        {
          or1k_cmp_op1 = force_reg(SImode, or1k_cmp_op1);
        }

      /* emit the set instruction. */
      tmp = gen_rtx(SET, VOIDmode, cc0_rtx,
                         gen_rtx(test, SImode,
                                 or1k_cmp_op0, or1k_cmp_op1));
      emit_insn(tmp);
  
      /* emit the branch instruction. */
      tmp = gen_rtx(SET, VOIDmode, 
                              pc_rtx,
                              gen_rtx(IF_THEN_ELSE, VOIDmode,
                                      gen_rtx(NE, VOIDmode, const0_rtx,
                                              cc0_rtx),
                                      gen_rtx(LABEL_REF,VOIDmode,operands[0]), 
                                      pc_rtx));
      emit_jump_insn( tmp);
      break;
    }

}

int
sym_ref_mem_operand(op, mode)
	register rtx op;
	enum machine_mode mode;
{
	if(GET_CODE(op) == MEM) {
		rtx t1 = XEXP(op, 0);
		if (GET_CODE(t1) == SYMBOL_REF)
			return 1;
	}

	return 0;
}

void function_prologue(file, vars)
	FILE *file;
	int vars;
{
	register int save_area;
	register int gpr_save_area = 0;
	register int lr_save_area;
	register int fp_save_area;
	register int stack_size;
	register int regno = 0;
	
	fprintf(file, "\n\t# ");

	for (regno = 0; regno < FIRST_PSEUDO_REGISTER; regno++) {
		fprintf(file,"%i", regs_ever_live[regno]);

		if (regs_ever_live[regno] && !call_used_regs[regno]) {
			gpr_save_area += 4;
		}
	}
	
	lr_save_area = !current_function_is_leaf || regs_ever_live[LINK_REGNUM] ? 4 : 0;
	fp_save_area = frame_pointer_needed ? 4 : 0;

	stack_size = OR1K_ALIGN(current_function_outgoing_args_size,4) + lr_save_area + fp_save_area + gpr_save_area + OR1K_ALIGN(vars,4);

	fprintf(file, "\n\t# gpr_save_area %d vars %d current_function_outgoing_args_size %d\n", gpr_save_area, vars, current_function_outgoing_args_size);

	if(stack_size >= 0x8000) {
		fprintf (file, "\tl.movhi   \tr%d,hi(%d)\n", GP_ARG_RETURN, stack_size);
                fprintf (file, "\tl.ori   \tr%d,r%d,lo(%d)\n", GP_ARG_RETURN, GP_ARG_RETURN, stack_size);
                fprintf (file, "\tl.sub   \tr%d,r%d,r%d\n", STACK_POINTER_REGNUM, STACK_POINTER_REGNUM, GP_ARG_RETURN);
	}
        else if(stack_size > 0)
                fprintf (file, "\tl.addi   \tr%d,r%d,%d\n", STACK_POINTER_REGNUM, STACK_POINTER_REGNUM, -stack_size);
	
	if (fp_save_area) {
		fprintf (file, "\tl.sw     \t%d(r%d),r%d\n", OR1K_ALIGN(current_function_outgoing_args_size,4) + lr_save_area, STACK_POINTER_REGNUM, FRAME_POINTER_REGNUM);
		if(stack_size >= 0x8000)
			fprintf (file, "\tl.add   \tr%d,r%d,r%d\n",  FRAME_POINTER_REGNUM, STACK_POINTER_REGNUM, GP_ARG_RETURN);
		else
			fprintf (file, "\tl.addi   \tr%d,r%d,%d\n",  FRAME_POINTER_REGNUM, STACK_POINTER_REGNUM, stack_size); 
	}

	if (lr_save_area)
        	fprintf (file, "\tl.sw     \t%d(r%d),r%d\n", OR1K_ALIGN(current_function_outgoing_args_size,4), STACK_POINTER_REGNUM, LINK_REGNUM);

	save_area = OR1K_ALIGN(current_function_outgoing_args_size,4) + lr_save_area + fp_save_area;
	for (regno = 0; regno < FIRST_PSEUDO_REGISTER; regno++)
		if (regs_ever_live[regno] && !call_used_regs[regno]) {
			fprintf (file, "\tl.sw    \t%d(r%d),r%d\n", save_area, STACK_POINTER_REGNUM, regno);
			save_area += 4;
		}
}

void function_epilogue(file, vars)
	FILE *file;
	int vars;
{
	register int save_area;
        register int gpr_save_area = 0;
        register int lr_save_area;
        register int fp_save_area;
        register int stack_size;
        register int regno = 0;

	for (regno = 0; regno < FIRST_PSEUDO_REGISTER; regno++)
                if (regs_ever_live[regno] && !call_used_regs[regno]) {
                        gpr_save_area += 4;
                }
 
	lr_save_area = !current_function_is_leaf || regs_ever_live[LINK_REGNUM] ? 4 : 0;
        fp_save_area = frame_pointer_needed ? 4 : 0;
 
        stack_size = OR1K_ALIGN(current_function_outgoing_args_size,4) + lr_save_area + fp_save_area + gpr_save_area + OR1K_ALIGN(vars,4);

	if (lr_save_area)
		fprintf (file, "\tl.lwz    \tr%d,%d(r%d)\n", LINK_REGNUM, OR1K_ALIGN(current_function_outgoing_args_size,4), STACK_POINTER_REGNUM);

	if (fp_save_area)
		fprintf (file, "\tl.lwz    \tr%d,%d(r%d)\n", FRAME_POINTER_REGNUM, OR1K_ALIGN(current_function_outgoing_args_size,4) + lr_save_area, STACK_POINTER_REGNUM);

	save_area = OR1K_ALIGN(current_function_outgoing_args_size,4) + lr_save_area + fp_save_area;
        for (regno = 0; regno < FIRST_PSEUDO_REGISTER; regno++)
                if (regs_ever_live[regno] && !call_used_regs[regno]) {
                        fprintf (file, "\tl.lwz    \tr%d,%d(r%d)\n", regno, save_area, STACK_POINTER_REGNUM);
                        save_area += 4;
                } 

	if(stack_size >= 0x8000) {
                fprintf (file, "\tl.movhi   \tr3,hi(%d)\n",  stack_size);
		fprintf (file, "\tl.ori   \tr3,r3,lo(%d)\n",  stack_size);
#if NOP_DELAY_SLOT_FILL
	        fprintf (file, "\tl.add   \tr%d,r%d,r3\n", STACK_POINTER_REGNUM, STACK_POINTER_REGNUM);
		fprintf (file, "\tl.jr  \tr%d\n", LINK_REGNUM);
		fprintf (file, "\tl.nop\n");
#else
		fprintf (file, "\tl.jr  \tr%d\n", LINK_REGNUM);
	        fprintf (file, "\tl.add   \tr%d,r%d,r3\n", STACK_POINTER_REGNUM, STACK_POINTER_REGNUM);
#endif
	}
	else if (stack_size > 0) {
#if NOP_DELAY_SLOT_FILL
		fprintf (file, "\tl.addi   \tr%d,r%d,%d\n", STACK_POINTER_REGNUM, STACK_POINTER_REGNUM, stack_size);
		fprintf (file, "\tl.jr  \tr%d\n", LINK_REGNUM);
		fprintf (file, "\tl.nop\n");
#else
		fprintf (file, "\tl.jr  \tr%d\n", LINK_REGNUM);
		fprintf (file, "\tl.addi   \tr%d,r%d,%d\n", STACK_POINTER_REGNUM, STACK_POINTER_REGNUM, stack_size);
#endif
	}
	else {
		fprintf (file, "\tl.jr  \tr%d\n", LINK_REGNUM);
		fprintf (file, "\tl.nop\n");
	}

	fprintf (file, ".endproc _%s\n", get_function_name());
}

int log_of_two(int x)
{
	int y;
  
	x >>= 1;
	for(y=0; x; x>>=1, y++);
      
	return y;
}

void
print_operand(file, x, code)
    FILE *file;
    rtx x;
    char code;
{ 									
  if (code == 'r' && GET_CODE (x) == MEM && GET_CODE (XEXP (x, 0)) == REG) 
    fprintf (file, "%s", reg_names[REGNO (XEXP (x, 0))]);		
  else if (code == '(')							
    if (dbr_sequence_length())						
      fprintf (file, "\t# delay slot filled");				
    else								
      fprintf (file, "\n\tl.nop\t\t\t# nop delay slot");			
  else if (code == 'C')							
     switch(GET_CODE(x))						
       {								
         case EQ:        fputs("eq", file); break;			
         case NE:        fputs("ne", file); break;			
         case GT:        fputs("gts", file); break;			
         case GE:        fputs("ges", file); break;			
         case LT:        fputs("lts", file); break;			
         case LE:        fputs("les", file); break;			
         case GTU:       fputs("gtu", file); break;			
         case GEU:       fputs("geu", file); break;			
         case LTU:       fputs("ltu", file); break;			
         case LEU:       fputs("leu", file); break;			
         default:							
           abort ();							
       }   								
  else if (GET_CODE (x) == REG)						
    fprintf (file, "%s", reg_names[REGNO (x)]);	
  else if (GET_CODE (x) == MEM)						
	output_address (XEXP (x, 0));					
  else 									
  { 									
	/*debug_rtx(x);*/ 						
	output_addr_const (file, x); 					
   }									
}

